#include <linux/module.h>
#include <linux/pci.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <asm/uaccess.h>
#include "sp_proto.h"
#include "spinmod.h"

MODULE_LICENSE("GPL");

int __amcc_outb(char data, unsigned short address);
int __pci_core_outb(char data, unsigned short address);

/* base address of board */
static unsigned long base_address;

/* major and minor numbers */
static dev_t maj_min_numbers;

/* cdev struct used to register the character device */
static struct cdev *sp_cdev;

ssize_t sp_read(struct file *f, char __user *user_buff, 
                size_t size, loff_t *offset) 
{
	DPRINTK("******************\n");
	DPRINTK("Enter sp_read\n");
	return 0;
}

ssize_t sp_write(struct file *f, const char __user *user_buff, 
                 size_t count, loff_t *offset) 
{
    int num_bytes; 
	struct sp_data data;

	DPRINTK("******************\n");
	DPRINTK("Enter sp_write\n");
	DPRINTK("count:%d\n", count);

	if ((num_bytes = copy_from_user(&data, user_buff, count)) != 0)
        return -1;

	DPRINTK("data:%d\n", data.data);
	DPRINTK("port:%d\n", data.port);

	__pci_core_outb(data.data, data.port);
	
	/* return count written */
	return count;
}

static struct file_operations sp_fops = {
	.owner = THIS_MODULE,
	.read = sp_read,
	.write = sp_write,
};

int spinmod_probe (struct pci_dev *dev, const struct pci_device_id *id)
{
	int ret = 0;
	DPRINTK("spinmod notified of board with vendorid=%x and deviceid=%x\n", 
            dev->vendor, dev->device);

    /* use dynamic allocation of major and minor numbers */
	if ((ret = alloc_chrdev_region(&maj_min_numbers, 0, 
                                   1, char_device_name))) {
		DPRINTK("spinmod couldn't get major and minor numbers\n");
		return ret;	
	}	

	printk("spinmod registered device named %s with \
major #: %d minor #: %d \n", 
            char_device_name, MAJOR(maj_min_numbers), MINOR(maj_min_numbers));
    
	/* alloc and fill out cdev struct */
	sp_cdev = cdev_alloc();
	sp_cdev->ops = &sp_fops;
    sp_cdev->owner = THIS_MODULE;
	
	if ((ret = cdev_add(sp_cdev, maj_min_numbers, 1) != 0)) {
        DPRINTK("call to cdev failed\n");
    }

    /* get base address of i/o region with 
     * base address register (bar) of 0 */
	base_address = pci_resource_start(dev, 0);
	return 0;
}

void spinmod_remove(struct pci_dev *dev) 
{
    /* unregister the character device */
	unregister_chrdev_region(maj_min_numbers, 1);

    /* delete the cdev */
	cdev_del(sp_cdev);	
}

static int __init init_func(void)
{
	DPRINTK("Calling pci_register_driver\n");
    /* alls we do is register the pci driver */
	return pci_register_driver(&spinmod_driver);
}


static void __exit exit_func(void)
{

	DPRINTK("Calling pci_unregister_driver\n");
    /* alls we do is unregister the pci driver 
     * spinmod_remove will be called from with the pci_device struct */
	pci_unregister_driver(&spinmod_driver);
}

module_init(init_func);
module_exit(exit_func);


/* the following are used by pbdfuncs.c */
/* write_word ()
 * Write a 4 byte word to the given IO port. Returns 0 on
 * success, -1 otherwise.
 */
int write_word (unsigned int data, unsigned int port)
{
  outl (data, base_address+ port);
  return 0;
}

/* write_byte ()
 * Write 1 byte to the given IO port. Returns 0 on success,
 * -1 otherwise.
 */
int write_byte (char data, unsigned int port)
{
  outb (data, base_address+ port);
  return 0;
}

/* read_word ()
 * Read 4 byte word from the given port.
 */
unsigned int read_word (unsigned int port)
{
  return inl (base_address+ port);
}

/* read_byte ()
 * Read 1 byte from the given port
 */
char read_byte (unsigned int port)
{
  return inb (base_address+ port);
}

/* write a byte to a board that uses the pci core */
int __pci_core_outb(char data, unsigned short address)
{
    write_byte(data, address); 
    return 0;
}


/* writes a byte to a board that uses the amcc chip.
   SP3 and previous revision boards use this interface. Later designs
   can be programmed directly. */
int __amcc_outb(char data, unsigned short address)
{
	unsigned int MAX_RECV_TIMEOUT = 10;
	unsigned int RECV_START, RECV_TOGGLE, RECV_TIMEOUT = 0;
	int XFER_ERROR = 0;
	int RECV_POLLING = 0;

	unsigned int OGMB = 0x0C;
	unsigned int CHK_RECV = 0x1F;
	unsigned int SIG_TRNS = 0x0F;

	unsigned int CLEAR31 = 0x00000001;
	unsigned int CLEAR24 = 0x000000FF;
	unsigned int CLEAR28 = 0x0000000F;
	unsigned int SET_XFER = 0x01000000;

	unsigned int Temp_Address = address;
	unsigned int Temp_Data = data;

		// Prepare Address Transfer
		Temp_Address &= CLEAR28;
		Temp_Address |= SET_XFER;

		// Prepare Data Transfer
		Temp_Data &= CLEAR24;
		Temp_Data |= SET_XFER;

		// Clear the XFER bit (Should already be cleared)
		write_byte (0,SIG_TRNS);
		mb();

		// Transfer Address
	
		// Read RECV bit from the Board
		RECV_START = read_byte(CHK_RECV);

		DPRINTK("RECV_START:%d\n", RECV_START);
		RECV_START &= CLEAR31;	// Only Save LSB


		// Send Address to OGMB
		write_word(Temp_Address, OGMB);
		mb();

		RECV_POLLING = 1;	// Set Polling Flag
		RECV_TIMEOUT = 0;
		while ((RECV_POLLING == 1) && (RECV_TIMEOUT < MAX_RECV_TIMEOUT)) 
		{
			RECV_TOGGLE = read_byte(CHK_RECV);
			mb();

			RECV_TOGGLE &= CLEAR31;	// Only Save LSB
			if (RECV_TOGGLE != RECV_START) RECV_POLLING = 0;	// Finished if Different
			else RECV_TIMEOUT++;
			if (RECV_TIMEOUT == MAX_RECV_TIMEOUT) XFER_ERROR = -2;
			DPRINTK("RECV_POLLING=%d\n", RECV_POLLING);
		}

		// Transfer Complete (Clear) Signal
		write_byte(0, SIG_TRNS);
		mb();

		// Transfer Data

		// Read RECV bit from the Board
		RECV_START = read_byte(CHK_RECV);
		DPRINTK("RECV_START:%d\n", RECV_START);
		
		RECV_START &= CLEAR31;	// Only Save LSB

		// Send Data to OGMB
		write_word(Temp_Data, OGMB);
		mb();

		RECV_POLLING = 1;	// Set Polling Flag
		RECV_TIMEOUT = 0;
		while ((RECV_POLLING == 1) && (RECV_TIMEOUT < MAX_RECV_TIMEOUT)) 
		{
			// Check for Toggled RECV
                        RECV_TOGGLE = read_byte(CHK_RECV);
			mb();

			DPRINTK("%d\n", RECV_TOGGLE);
			RECV_TOGGLE &= CLEAR31;	// Only Save LSB
			if (RECV_TOGGLE != RECV_START) RECV_POLLING = 0;	// Finished if Different
			else RECV_TIMEOUT++;
			if (RECV_TIMEOUT == MAX_RECV_TIMEOUT) XFER_ERROR = -2;
		}

		// Transfer Complete (Clear) Signal
		write_byte(0, SIG_TRNS);


	return XFER_ERROR;
	
}

